/*
uptimed - Copyright (c) 1998-2000 Rob Kaper <cap@capsi.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 dated June, 1991.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program;  if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave., Cambridge, MA 02139, USA.
*/

#include "uptimed.h"

#ifdef PLATFORM_HPUX
#define	SYSLOG_PREFIX	"uptimed: "
#else
#define SYSLOG_PREFIX	""
#endif

Urec *u_current;
int our_pos;
int update_interval=60, max_entries=50, create_bootid=0, send_email=0;
char email[EMAIL+1], sendmail[EMAIL+1];
       
int main(int argc, char *argv[])
{
	Milestone *next_milestone, *cheer_milestone;
	time_t prev=0;
	int cheer_pos=0;
	void handler(int);

	/* Read configuration settings. */
	read_config();

	/* Scan arguments. */
	scan_args(argc, argv);
			
	/* Create bootid and exit. Should be done once on startup. */
	if (create_bootid)
	{
		createbootid();
		syslog(LOG_INFO, SYSLOG_PREFIX "created bootid: %d", readbootid());
		exit(0);
	}

	if (update_interval<1)
		update_interval=1;

	if (max_entries<0)
		max_entries=1;

	/* Let's start with adding a nice entry for the current session. */
	u_current=add_urec(read_uptime(),readbootid(),read_sysinfo());

	/* How about adding the existing records as well? */
	read_records(u_current->btime);

	/* Initialize the first milestone to look for. */
	next_milestone = find_next_milestone(u_current->utime);

 	/* Add signal handler for hangups/terminators. */
 #ifndef PLATFORM_HPUX
 	signal(SIGHUP,handler);
 #endif
 	signal(SIGTERM,handler);
 
	/* Now run in the background. */
	bg();

	/* The main loop. */
	while (1)
	{
		/* Update the uptime value of the current session. */
		u_current->utime=read_uptime();

		/* Find out if we're breaking existing records. */
		if (u_current!=urec_list)
		{
			if (!prev) prev=get_prev();

			/* Current uptime might break the old record. */
			if (u_current->utime > urec_list->utime)
			{
				moveup(); /* We should be urec_list ourselves now. */
				cheer_pos=1; /* Don't forget to cheer. */
			}
			/* Current uptime might move up in the list. */
			else if (prev && u_current->utime > prev)
			{
				moveup();
				prev=get_prev(); /* We will have to re-init the prev entry. */
				cheer_pos=our_pos; /* Don't forget to cheer. */
			}
		}

		/* Find out if we are breaking any milestones. */
		if (next_milestone && u_current->utime > next_milestone->time)
		{
			cheer_milestone = next_milestone;
			next_milestone = find_next_milestone(u_current->utime);
		}

		/* Celebrate good times! */		 
		if (cheer_pos || cheer_milestone)
		{
			cheer(cheer_milestone, cheer_pos);
			cheer_pos = 0;
			cheer_milestone = NULL;
		}
		
		/* Save all the records. */
		save_records(max_entries);

		/* Save valueable CPU cycles. */
		sleep(update_interval);
	}
}

void bg(void)
{
	/* Simple fork to run proces in the background. */
	switch(fork())
	{
		case 0: case -1:
			break;
		default:
			exit(0);
	}
}

void cheer(Milestone *milestone, int position)
{
	/* This function will do various sorts of recordbreaking celebrating. */
	switch(position)
	{
		case 0:
			break;
		case 1:
			syslog(LOG_INFO, SYSLOG_PREFIX "new uptime record: %s", time2uptime(u_current->utime));
			break;
		default:
			syslog(LOG_INFO, SYSLOG_PREFIX "moving up to position %d: %s", position, time2uptime(u_current->utime));
			break;
	}
	
	if (milestone!=NULL && milestone->time>0)
		syslog(LOG_INFO, SYSLOG_PREFIX "milestone: %s (%s)", time2uptime(milestone->time), milestone->desc);

	/* Send email if it's requested. */
	if (send_email && email[0])
		mail(milestone, position);
}

time_t get_prev(void)
{
	Urec *uprev=NULL, *u;

	our_pos=0;
	for (u=urec_list;u;u=u->next)
	{
		our_pos++;
		if (u==u_current)
			if (uprev)
				return uprev->utime;
		uprev=u;
	}
	return 0;
}

void read_config(void)
{
	FILE *f;
	char str[256];
	time_t milestone_time;
	char *milestone_str;
	
	f=fopen(FILE_CONFIG, "r");
	if (!f) return;
		
	fgets(str, sizeof(str), f);
	while (!feof(f))
	{
		if (!strncmp(str, "UPDATE_INTERVAL",  15))
			update_interval=atoi(str+16);
		else if (!strncmp(str, "MAX_ENTRIES",  11))
			max_entries=atoi(str+12);
		else if (!strncmp(str, "EMAIL",  5) && strstr(str+6, "@"))
		{
			strncpy(email, strtok(str+6, "\n"), EMAIL);
			email[EMAIL]='\0';
		}
		else if (!strncmp(str, "SEND_EMAIL",  10))
			send_email=atoi(str+11);
		else if (!strncmp(str, "SENDMAIL",  8))
		{
			strncpy(sendmail, strtok(str+9, "\n"), EMAIL);
			sendmail[EMAIL]='\0';
		}
		else if (!strncmp(str, "MILESTONE",  9))
		{
			char	*cp;
			char	*end;
			size_t	 len;
			time_t	 multiplier;

			/* Find the last char in string. */
			cp = strtok(str+10, ":");
			len = strlen(cp);
			end = cp + len - 1;

			/* And examine it. */
 			if (isdigit(*end))
 			{
 				/* It's a digit, so input was in seconds. */
 				multiplier = 1;
			}
			else
			{
				switch(tolower(*end))
				{
					case 's':
						/* Seconds. */
						multiplier = 1;
						break;

					case 'd':
						/* Days. */
						multiplier = 86400;
						break;

					case 'w':
						/* Weeks. */
						multiplier = 7 * 86400;
						break;

					case 'y':
						/* Solar years. */
						multiplier = (time_t) (365.24219 * 86400.00);
						break;

					default:
						/* Garbage. */
						multiplier = 0;
						break;
				}
				*end = '\0';		/* remove the time factor byte. */
			}
			milestone_time=atol(cp)*multiplier;
			milestone_str=strtok(NULL, "\n");
			add_milestone(milestone_time, milestone_str);
		}
		
		fgets(str, sizeof(str), f);
	}
}

/*
Handler does not care about sorting the list or cheering, there's no time
for all that when we've received a SIGHUP or SIGTERM.

When uptimed is restarted the list will be sorted automatically and we'll
try not to shed any tears over any missing syslog messages that might have
been printed at the end of the interval.
*/
void handler(int s)
{
	/* Update the uptime value of the current session. */
	u_current->utime=read_uptime();

	/* Save all the records. */
	save_records(max_entries);

	/* Exit gracefully. */
	exit(0);
}

void scan_args(int argc, char *argv[])
{
	int i=1;

	for (;i<argc;i++)
	{
		switch(argv[i][0])
		{
			case '-':
				switch(argv[i][1])
				{
					case '?':
					case 'e':
					case 'v':
						if (strchr(argv[i], '?'))
							print_help();
						if (strchr(argv[i], 'v'))
							print_version();
						if (strchr(argv[i], 'e'))
						{
							if (argv[i]+2 && strstr(argv[i]+2, "@"))
							{
								strncpy(email, argv[i]+2, EMAIL);
								email[EMAIL]='\0';
							}
							else if (argc>i+1 && strstr(argv[i+1], "@"))
							{
								strncpy(email, argv[++i], EMAIL);
								email[EMAIL]='\0';
							}
							send_email++;
						}
						break;
					case 'b': /* ignore the rest, -boot will still work */
						create_bootid++;
						break;
					case 'i':
						if (isdigit(*(argv[i]+2)))
							update_interval=atoi(argv[i]+2);
						else if (argc>i+1 && isdigit(*argv[i+1]))
							update_interval=atoi(argv[++i]);
						else
						{
							printf("%s: option requires argument -- %s\n", argv[0], argv[i]+1);
							print_usage();
						}
						break;
					case 'm':
						if (isdigit(*(argv[i]+2)))
							max_entries=atoi(argv[i]+2);
						else if (argc>i+1 && isdigit(*argv[i+1]))
							max_entries=atoi(argv[++i]);
						else
						{
							printf("%s: option requires argument -- %s\n", argv[0], argv[i]+1);
							print_usage();
						}
						break;
					default:
						printf("%s: invalid option -- %s\n", argv[0], argv[i]+1);
						print_usage();
						break;
				}
				break;
			default:
				break;
		}
	}
}

void print_usage(void)
{
	printf("usage: uptimed [-?bv] [-e email] [-i interval] [-m count]\n");
	exit(1);
}

void print_help(void)
{
	printf("usage: uptimed [OPTION]...\n");
	printf("commandline options override settings from configuration file\n\n");
	printf("  -?             this help\n");
	printf("  -b             create bootid and exit\n");
	printf("  -e EMAIL       send mail to EMAIL at milestones/records\n");
	printf("  -i INTERVAL    use INTERVAL seconds for loop\n");
	printf("  -m COUNT       log a maximum of COUNT entries\n");
	printf("  -v             version information\n");
	exit(0);
}

void print_version(void)
{
	printf("uptimed " VERSION " by Rob Kaper <cap@capsi.com>\n");
	exit(0);
}

/* Ugly mail hack, assumes sendmail compatible mail agent. */
void mail(Milestone *milestone, int position)
{
	char buf[256], str[2048], cmd[2048];

	sprintf(str, "To: %s\r\nSubject: Congratulations\r\n\r\n", email);

	if (milestone)
	{
		sprintf(buf, "Your uptime has reached another milestone: %s (%s)\r\n", time2uptime(milestone->time), milestone->desc);
		strcat(str, buf);
	}

	if (position)
	{
		if (position==1)
			sprintf(buf, "Your uptime is a record for this system: %s\r\n", time2uptime(u_current->utime));
		else
			sprintf(buf, "Your uptime moved up to position %d: %s\r\n", position, time2uptime(u_current->utime));
		strcat(str, buf);
	}

	strcat(str, "\r\nCongratulations!\r\r\nUptimed author,\r\n\r\nRob Kaper <cap@capsi.com>\r\n-- \r\nThis message was automatically generated by uptimed.\r\n");

	sprintf(cmd, "printf \"%s\" | %s", str, sendmail);	
	system(cmd);
}
